import * as mc from "@minecraft/server";
import { config } from "../config.js";
import * as db from "../database.js";

var beforeItemUseOn = {};
var selection = {};
const vectors = [[0, 1, 0], [0, -1, 0], [0, 0, 1], [0, 0, -1], [1, 0, 0], [-1, 0, 0]];
const size_limit = (config.area_size_limit < 32768) ? config.area_size_limit : 32768;
const raycast_length = 64;
const areas = new db.database('bm_areas')

mc.world.events.worldInitialize.subscribe(({ propertyRegistry }) => {
    const worldPropertiesDefinition = new mc.DynamicPropertiesDefinition();
    worldPropertiesDefinition.defineString("areas", 9512);
    worldPropertiesDefinition.defineBoolean('bm_initialized');
    propertyRegistry.registerWorldDynamicProperties(worldPropertiesDefinition);
    if (!(mc.world.getDynamicProperty('bm_initialized'))) {
        mc.world.getDimension("minecraft:overworld").runCommandAsync(`gamerule sendCommandFeedback false`)
        mc.world.setDynamicProperty('bm_initialized', true)
    }
    if (mc.world.getDynamicProperty('areas') !== undefined && mc.world.getDynamicProperty('areas') !== 'updated') {
        areas.set(JSON.parse(mc.world.getDynamicProperty('areas')))
        mc.world.setDynamicProperty('areas', 'updated')
    }
});

function pack(pos, floor = false) {
    if (floor) {
        return [Math.floor(pos.x), Math.floor(pos.y), Math.floor(pos.z)];
    } else {
        return [pos.x, pos.y, pos.z];
    }
}

function addArea(pos1, pos2, dim) {
    let a = areas.get();
    let pos1_final = [Math.min(pos1[0], pos2[0]), Math.min(pos1[1], pos2[1]), Math.min(pos1[2], pos2[2])]
    let pos2_final = [Math.max(pos1[0], pos2[0]), Math.max(pos1[1], pos2[1]), Math.max(pos1[2], pos2[2])]
    a.push({ p1: pos1_final, p2: pos2_final, m: [0, 0, 0, 0, 0, 0], s: 1, d: dim });
    areas.set(a);
}

function within(pos, area) {
    let pos1 = area.p1;
    let pos2 = area.p2;
    return (
        pos[0] >= pos1[0] &&
        pos[0] <= pos2[0] &&
        pos[1] >= pos1[1] &&
        pos[1] <= pos2[1] &&
        pos[2] >= pos1[2] &&
        pos[2] <= pos2[2]);
}

function intersect(a1, a2) {
    let ap1 = a1.p1;
    let ap2 = a1.p2;
    let bp1 = a2.p1;
    let bp2 = a2.p2;
    return (
        ap1[0] < bp2[0] &&
        ap2[0] > bp1[0] &&
        ap1[1] < bp2[1] &&
        ap2[1] > bp1[1] &&
        ap1[2] < bp2[2] &&
        ap2[2] > bp1[2]
    );
}

async function render(a, p) {
    if (config.disable_render) {
        return true;
    }
    let pos = [a.p1, a.p2];
    let size;
    let density;
    if (config.dynamic_density) {
        size = (pos[1][0] - pos[0][0]) * (pos[1][1] - pos[0][1]) * (pos[1][2] - pos[0][2]);
        if (config.fixed_particle_amount > 0) {
            density = 1 / ((pos[1][0] - pos[0][0]) / config.fixed_particle_amount);
        } else {
            density = config.density - (size) / config.selection_size_limit * config.density;
        }
    } else {
        if (config.fixed_particle_amount > 0) {
            density = 1 / ((pos[1][0] - pos[0][0]) / config.fixed_particle_amount);
        } else {
            density = config.density
        }
    }
    let console = mc.world.getDimension(a.d)
    await console.runCommandAsync(`particle ${p} ${pos[0][0]} ${pos[0][1] + 0.5} ${pos[0][2]}`);
    let commands = 0
    for (let y = 0; y < 2; y++) {
        for (let z = 0; z < 2; z++) {
            for (let i = 0; i <= (pos[1][0] - pos[0][0]) * density; i++) {
                if (commands < 100) {
                    console.runCommandAsync(`particle ${p} ${pos[0][0] + i / density} ${pos[y][1] + 0.5} ${pos[z][2]}`);
                    commands++;
                } else {
                    commands = 0;
                    await console.runCommandAsync(`particle ${p} ${pos[0][0] + i / density} ${pos[y][1] + 0.5} ${pos[z][2]}`);
                }
            }
        }
    }
    if (config.fixed_particle_amount > 0) {
        density = 1 / ((pos[1][2] - pos[0][2]) / config.fixed_particle_amount);
    }
    for (let y = 0; y < 2; y++) {
        for (let x = 0; x < 2; x++) {
            for (let i = 0; i <= (pos[1][2] - pos[0][2]) * density; i++) {
                if (commands < 100) {
                    console.runCommandAsync(`particle ${p} ${pos[x][0]} ${pos[y][1] + 0.5} ${pos[0][2] + i / density}`);
                    commands++;
                } else {
                    commands = 0;
                    await console.runCommandAsync(`particle ${p} ${pos[x][0]} ${pos[y][1] + 0.5} ${pos[0][2] + i / density}`);
                }
            }
        }
    }
    if (config.fixed_particle_amount > 0) {
        density = 1 / ((pos[1][1] - pos[0][1]) / config.fixed_particle_amount);
    }
    for (let x = 0; x < 2; x++) {
        for (let z = 0; z < 2; z++) {
            for (let i = 0; i <= (pos[1][1] - pos[0][1]) * density; i++) {
                if (commands < 100) {
                    console.runCommandAsync(`particle ${p} ${pos[x][0]} ${pos[0][1] + i / density + 0.5} ${pos[z][2]}`);
                    commands++;
                } else {
                    commands = 0;
                    await console.runCommandAsync(`particle ${p} ${pos[x][0]} ${pos[0][1] + i / density + 0.5} ${pos[z][2]}`);
                }
            }
        }
    }
}

mc.world.events.beforeItemUse.subscribe((evt) => {
    let item = evt.item;
    let name = evt.source.name;
    if (item.typeId == 'blockmotion:wrench') {
        if (selection[name] == undefined) {
            if (evt.source.isSneaking) {
                let a = areas.get();
                for (let i = 0; i < a.length; i++) {
                    if (raycast(evt.source, a[i])) {
                        a.splice(i, 1);
                        evt.source.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.deleted" } ] }`)
                        areas.set(a);
                        break;
                    }
                }
            } else {
                selection[name] = pack(evt.source.getBlockFromViewDirection());
                evt.source.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.pos1" } ] }`)
            }
        } else {
            let collision = false;
            let pos1 = selection[name]
            let pos2 = pack(evt.source.getBlockFromViewDirection({ maxDistance: 128 }))
            let asp1 = [Math.min(pos1[0], pos2[0]), Math.min(pos1[1], pos2[1]), Math.min(pos1[2], pos2[2])]
            let asp2 = [Math.max(pos1[0], pos2[0]), Math.max(pos1[1], pos2[1]), Math.max(pos1[2], pos2[2])]
            let areaSelection = { p1: asp1, p2: asp2 };
            let a = areas.get();
            let size = (asp2[0] - asp1[0]) * (asp2[1] - asp1[1]) * (asp2[2] - asp1[2]);
            for (let i = 0; i < a.length; i++) {
                if (intersect(areaSelection, a[i])) {
                    collision = true;
                }
            }
            if (collision) {
                evt.source.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.intersect" } ] }`)
            } else if (size < size_limit) {
                addArea(pos1, pos2, evt.source.dimension.id);
                evt.source.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.connected" } ] }`);
                delete selection[name];
            } else {
                evt.source.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.too_big" } ] }`)
            }
        }
    }
})

mc.world.events.tick.subscribe(async (evt) => {
    if ((evt.currentTick % config.render_update_rate) === 0) {
        for (let player of mc.world.getPlayers()) {
            let selectedItem = player.getComponent("inventory").container.getItem(player.selectedSlot);
            let selectedItemId = (selectedItem == undefined) ? 'minecraft:air' : selectedItem.typeId;
            if (Object.keys(selection).includes(player.name)) {
                if (selectedItemId != 'blockmotion:wrench' || player.isSneaking) {
                    player.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.cancelled" } ] }`)
                    delete selection[player.name];
                } else {
                    let collision = false;
                    let pos1 = selection[player.name]
                    let pos2 = pack(player.getBlockFromViewDirection({ maxDistance: 128 }))
                    let asp1 = [Math.min(pos1[0], pos2[0]), Math.min(pos1[1], pos2[1]), Math.min(pos1[2], pos2[2])]
                    let asp2 = [Math.max(pos1[0], pos2[0]), Math.max(pos1[1], pos2[1]), Math.max(pos1[2], pos2[2])]
                    let areaSelection = { p1: asp1, p2: asp2, d: player.dimension.id };
                    let a = areas.get();
                    for (let area of a) {
                        let distance = [Math.abs(player.location.x - (area.p1[0] + area.p2[0]) / 2), Math.abs(player.location.y - (area.p2[1] + area.p1[1]) / 2), Math.abs(player.location.z - (area.p2[2] + area.p1[2]) / 2)];
                        if (intersect(areaSelection, area)) {
                            collision = true;
                            await render(area, 'blockmotion:red');
                        } else if (distance[0] < 64 && distance[1] < 64 && distance[2] < 64) {
                            await render(area, 'blockmotion:green');
                        }
                    }
                    let size = (asp2[0] - asp1[0]) * (asp2[1] - asp1[1]) * (asp2[2] - asp1[2]);
                    if (!(collision)) {
                        if (size < size_limit) {
                            await render(areaSelection, 'blockmotion:green');
                        } else {
                            if (config.render_big_areas) {
                                await render(areaSelection, 'blockmotion:yellow');
                            }
                            player.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.too_big" } ] }`)
                        }
                    }
                }
            } else if (selectedItemId == 'blockmotion:wrench') {
                let a = areas.get();
                if (player.isSneaking) {
                    for (let area of a) {
                        let distance = [Math.abs(player.location.x - (area.p1[0] + area.p2[0]) / 2), Math.abs(player.location.y - (area.p2[1] + area.p1[1]) / 2), Math.abs(player.location.z - (area.p2[2] + area.p1[2]) / 2)];
                        if (distance[0] < 64 && distance[1] < 64 && distance[2] < 64) {
                            if (raycast(player, area)) {
                                await render(area, 'blockmotion:red');
                            } else {
                                await render(area, 'blockmotion:white');
                            }
                        }
                    }
                } else {
                    for (let area of a) {
                        let distance = [Math.abs(player.location.x - (area.p1[0] + area.p2[0]) / 2), Math.abs(player.location.y - (area.p2[1] + area.p1[1]) / 2), Math.abs(player.location.z - (area.p2[2] + area.p1[2]) / 2)];
                        if (distance[0] < 64 && distance[1] < 64 && distance[2] < 64) {
                            await render(area, 'blockmotion:white');
                        }
                    }
                }
            }
        }
    }
})

function raycast(player, area) {
    let vector = player.viewDirection;
    for (let i = 0; i < raycast_length; i++) {
        if (within([
            player.headLocation.x + i * vector.x,
            player.headLocation.y + i * vector.y,
            player.headLocation.z + i * vector.z,
        ], area)) {
            return true;
        }
    }
    return false;
}

function findArea(block) {
    let a = areas.get();
    return a.findIndex((e) => {
        if (within(pack(block), e)) {
            return true;
        }
    })
}

mc.world.events.beforeItemUseOn.subscribe((evt) => {
    if (beforeItemUseOn[evt.source.name] > Date.now()) {
        return;
    }
    beforeItemUseOn[evt.source.name] = Date.now() + 250;
    let player = evt.source;
    let block = player.dimension.getBlock(evt.blockLocation);
    if (block.typeId == 'blockmotion:controls') {
        let a = areas.get();
        let area = a[findArea(block)];
        let pos1 = area.p1;
        let pos2 = area.p2;
        if (evt.item.typeId != 'minecraft:wooden_hoe' &&
            evt.item.typeId != 'minecraft:stone_hoe' &&
            evt.item.typeId != 'minecraft:iron_hoe' &&
            evt.item.typeId != 'minecraft:golden_hoe' &&
            evt.item.typeId != 'minecraft:diamond_hoe' &&
            evt.item.typeId != 'minecraft:netherite_hoe'
        ) {
            if (block.permutation.getProperty('blockmotion:speed').value) {
                let speed = area.s;
                if (player.isSneaking) {
                    if (speed > config.minimum_speed) {
                        speed -= 0.2;
                        player.runCommandAsync(`titleraw @a actionbar { "rawtext": [ {"text": "§c${Math.round(speed * 10) / 10} "}, { "translate": "text.bps" } ] }`)
                        player.runCommandAsync(`playsound random.click @a ${block.x} ${block.y} ${block.z} 0.2 ${2 * (speed / 5)}`)
                    } else {
                        player.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.minimal_speed" } ] }`)
                        player.runCommandAsync(`playsound note.bass @a ${block.x} ${block.y} ${block.z} 0.4`)
                    }
                } else {
                    if (speed < config.maximum_speed) {
                        speed += 0.2;
                        player.runCommandAsync(`titleraw @a actionbar { "rawtext": [ {"text": "§a${Math.round(speed * 10) / 10} "}, { "translate": "text.bps" } ] }`)
                        player.runCommandAsync(`playsound random.click @a ${block.x} ${block.y} ${block.z} 0.2 ${2 * (speed / 5)}`)
                    } else {
                        player.runCommandAsync(`titleraw @a actionbar { "rawtext": [ { "translate": "text.maximal_speed" } ] }`)
                        player.runCommandAsync(`playsound note.bass @a ${block.x} ${block.y} ${block.z} 0.4`)
                    }
                }
                area.s = speed;
            } else {
                let permutation = block.permutation;
                let turn = permutation.getProperty('blockmotion:turn');
                turn.value = !(turn.value);
                block.setPermutation(permutation);
                area.m[block.permutation.getProperty('blockmotion:rotation').value] = (turn.value) ? 1 : 0;
                if (turn.value) {
                    player.runCommandAsync(`playsound random.click @a ${block.x} ${block.y} ${block.z} 0.3 0.6`)
                } else {
                    player.runCommandAsync(`playsound random.click @a ${block.x} ${block.y} ${block.z} 0.3 0.5`)
                }
            }
            areas.set(a);
        } else {
            let permutation = block.permutation;
            let speed = permutation.getProperty('blockmotion:speed');
            speed.value = !(speed.value);
            block.setPermutation(permutation);
            if (speed.value) {
                player.runCommandAsync(`playsound open.fence_gate @a ${block.x} ${block.y} ${block.z} 0.3`)
            } else {
                player.runCommandAsync(`playsound close.fence_gate @a ${block.x} ${block.y} ${block.z} 0.3`)
            }
        }
    }
})

mc.world.events.blockBreak.subscribe((evt) => {
    if (evt.brokenBlockPermutation.type.id == 'blockmotion:controls') {
        let a = areas.get();
        let area = a[findArea(evt.block)];
        area.m[evt.brokenBlockPermutation.getProperty('blockmotion:rotation').value] = 0;
        areas.set(a)
    }
})

function testAir(area, move, idx) {
    let pos1 = area.p1;
    let pos2 = area.p2;
    let sizeX = pos2[0] - pos1[0];
    let sizeY = pos2[1] - pos1[1];
    let sizeZ = pos2[2] - pos1[2];

    if (idx == 0) {
        if (move[0] < 0) {
            for (let z = 0; z <= sizeZ; z++) {
                for (let y = 0; y <= sizeY; y++) {
                    if (mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] - 1, pos1[1] + y, pos1[2] + z)).type !== mc.MinecraftBlockTypes.air && mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] - 1, pos1[1] + y, pos1[2] + z)).type !== mc.MinecraftBlockTypes.water) {
                        return false;
                    }
                }
            }
        } else if (move[0] > 0) {
            for (let z = 0; z <= sizeZ; z++) {
                for (let y = 0; y <= sizeY; y++) {
                    if (mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos2[0] + 1, pos1[1] + y, pos1[2] + z)).type !== mc.MinecraftBlockTypes.air && mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos2[0] + 1, pos1[1] + y, pos1[2] + z)).type !== mc.MinecraftBlockTypes.water) {
                        return false;
                    }
                }
            }
        }
    } else if (idx == 1) {
        if (move[1] < 0) {
            for (let x = 0; x <= sizeX; x++) {
                for (let z = 0; z <= sizeZ; z++) {
                    if (mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos1[1] - 1, pos1[2] + z)).type !== mc.MinecraftBlockTypes.air && mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos1[1] - 1, pos1[2] + z)).type !== mc.MinecraftBlockTypes.water) {
                        return false;
                    }
                }
            }
        } else if (move[1] > 0) {
            for (let x = 0; x <= sizeX; x++) {
                for (let z = 0; z <= sizeZ; z++) {
                    if (mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos2[1] + 1, pos1[2] + z)).type !== mc.MinecraftBlockTypes.air && mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos2[1] + 1, pos1[2] + z)).type !== mc.MinecraftBlockTypes.water) {
                        return false;
                    }
                }
            }
        }
    } else if (idx == 2) {
        if (move[2] < 0) {
            for (let x = 0; x <= sizeX; x++) {
                for (let y = 0; y <= sizeY; y++) {
                    if (mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos1[1] + y, pos1[2] - 1)).type !== mc.MinecraftBlockTypes.air && mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos1[1] + y, pos1[2] - 1)).type !== mc.MinecraftBlockTypes.water) {
                        return false;
                    }
                }
            }
        } else if (move[2] > 0) {
            for (let x = 0; x <= sizeX; x++) {
                for (let y = 0; y <= sizeY; y++) {
                    if (mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos1[1] + y, pos2[2] + 1)).type !== mc.MinecraftBlockTypes.air && mc.world.getDimension(area.d).getBlock(new mc.BlockLocation(pos1[0] + x, pos1[1] + y, pos2[2] + 1)).type !== mc.MinecraftBlockTypes.water) {
                        return false;
                    }
                }
            }
        }
    }
    return true;
}

function add(d1, d2) {
    return [d1[0] + d2[0], d1[1] + d2[1], d1[2] + d2[2]];
}

async function clearBlocks(area, move) {
    let pos1 = area.p1;
    let pos2 = area.p2;
    if (move[0] < 0) {
        mc.world.getDimension(area.d).runCommandAsync(`fill ${pos2[0]} ${pos1[1]} ${pos1[2]} ${pos2[0] + (move[0] + 1)} ${pos2[1]} ${pos2[2]} air`)
    } else if (move[0] > 0) {
        mc.world.getDimension(area.d).runCommandAsync(`fill ${pos1[0]} ${pos1[1]} ${pos1[2]} ${pos1[0] + (move[0] - 1)} ${pos2[1]} ${pos2[2]} air`)
    }
    if (move[1] < 0) {
        mc.world.getDimension(area.d).runCommandAsync(`fill ${pos1[0]} ${pos2[1]} ${pos1[2]} ${pos2[0]} ${pos2[1] + (move[1] + 1)} ${pos2[2]} air`)
    } else if (move[1] > 0) {
        mc.world.getDimension(area.d).runCommandAsync(`fill ${pos1[0]} ${pos1[1]} ${pos1[2]} ${pos2[0]} ${pos1[1] + (move[1] - 1)}  ${pos2[2]} air`)
    }
    if (move[2] < 0) {
        mc.world.getDimension(area.d).runCommandAsync(`fill ${pos1[0]} ${pos1[1]} ${pos2[2]} ${pos2[0]} ${pos2[1]} ${pos2[2] + (move[2] + 1)} air`)
    } else if (move[2] > 0) {
        mc.world.getDimension(area.d).runCommandAsync(`fill ${pos1[0]} ${pos1[1]} ${pos1[2]} ${pos2[0]} ${pos2[1]} ${pos1[2] + (move[2] - 1)} air`)
    }
}

async function movePlayers(area, move) {
    let pos1 = area.p1;
    let pos2 = area.p2;
    mc.world.getDimension(area.d).runCommandAsync(`execute as @e[x=${pos1[0]},y=${pos1[1]},z=${pos1[2]},dx=${pos2[0] - pos1[0]},dy=${pos2[1] - pos1[1]},dz=${pos2[2] - pos1[2]}] positioned as @s run tp @s ~${move[0]} ~${move[1]} ~${move[2]}`);
    return true;
}

async function moveArea(area, move) {
    let pos1 = area.p1;
    let pos2 = area.p2;
    await mc.world.getDimension(area.d).runCommandAsync(`clone ${pos1[0]} ${pos1[1]} ${pos1[2]} ${pos2[0]} ${pos2[1]} ${pos2[2]} ${pos1[0] + move[0]} ${pos1[1] + move[1]} ${pos1[2] + move[2]} replace force`);
    return true;
}


mc.world.events.tick.subscribe((evt) => {
    let a = areas.get();
    for (let area of a) {
        if ((evt.currentTick % Math.round(20 / area.s)) === 0) {
            let enabled = false;
            area.m.forEach((e) => {
                if (e == 1) {
                    enabled = true;
                }
            })
            if (enabled) {
                let move = [0, 0, 0];
                let pos1 = area.p1;
                let pos2 = area.p2;
                for (let dir = 0; dir < 6; dir++) {
                    if (area.m[dir] != 0) {
                        move = add(move, vectors[dir])
                    }
                }
                for (let i = 0; i < 3; i++) {
                    if (!testAir(area, move, i)) {
                        move[i] = 0;
                    }
                }
                if (pos2[1] >= 319 && move[1] > 0) {
                    move[1] = 0;
                }
                let isMoving = (move != [0, 0, 0])
                if (isMoving) {
                    moveArea(area, move).then((r) => {
                        movePlayers(area, move);
                        clearBlocks(area, move);
                        area.p1 = add(pos1, move)
                        area.p2 = add(pos2, move)
                        areas.set(a)
                    })

                }
            }
        }
    }
})